# shellcheck disable=SC2034
LUM_CFG_FILE="/etc/ldapusermgmt/ldapusermgmt.conf"

# usage <retval> <synopsys> [..]
function usage() {
    (( $# > 1 )) || \
	bb_fatal "common:usage called with $# instead of 2 or more args"

    local _retval="$1"
    local _hint="$2"
    shift 2

    echo "${0##*/}: usage: ${0##*/} [OPTIONS] $1" >&2
    (( $# == 1 )) || {
	local _synopsys
	local _padding="${0##*/}: usage:"

	shift
	for _synopsys in "$@"
	do
	    printf "%${#_padding}s %s [OPTIONS] %s\n" \
		"" "${0##*/}" "$_synopsys" >&2
	done
    }
    cat >&2 <<EOF
  where options are:
    -c <config-file>                           .. location of the config file
    -h                                                 .. show this help text
    -F                              .. force operation, never ask (dangerous)
    -s <key>                           .. work on that section of config file
    -l                      .. work only on LDAP DB, skip operations on files
    -n                                      .. dry run, do not modify LDAP DB
    -q                                                  ,, decrease verbosity
    -v                                                  .. increase verbosity
    -V                                                 .. prints version info
EOF

    # write out local help
    cat >&2

    [ -z "$_hint" ] || printf "Usage hint: %s\n" "$_hint"

    bb_quit "$_retval"
}

# verbosity_init [<type>]
# the default type is stderr
# shellcheck disable=SC2120
function verbosity_init() {
    (( $# == 0 || $# == 1 )) || \
	bb_fatal "common:verbosity_init called with $# instead of 0 or 1 arg"

    local _type="${1-stderr}"

    declare -g VERBOSITY=2
    bb_msg_clear_targets
    bb_msg_set_target log_notrace "${_type}" info:alert prog notrace
    bb_msg_set_target log_trace "${_type}" emerg prog trace
}

# verbosity_set <verbosity> [<type>]
function verbosity_set() {
    (( $# == 1 || $# == 2 )) || bb_fatal "common:verbosity_set" \
	"called with $# args instead of 1 or 2 args"
    bb_validate_int "common:verbosity_set" "$1"

    local _type="${2-stderr}"

    local -A _levels=([0]="err" [1]="warning" [2]="info" [3]="debug")
    
    VERBOSITY="$1"
    (( VERBOSITY >= 0 )) || VERBOSITY=0
    (( VERBOSITY <= 4 )) || VERBOSITY=4

    if (( VERBOSITY == 4 ))
    then
	bb_msg_set_target log_notrace "${_type}" debug:alert full notrace
	bb_msg_set_target log_trace "${_type}" emerg full trace
    else
	bb_msg_set_target log_notrace "${_type}" \
	    "${_levels[$VERBOSITY]}:alert" prog notrace
	bb_msg_set_target log_trace "${_type}" emerg prog trace
    fi
}

function bb_msg_from_stdin() {
    (( $# == 2 )) || {
	bb_ierr "bb_msg_from_stdin called with $# instead of 2 args"
	return 2
    }

    if [ -t 0 ]
    then
	return
    fi

    local _line
    local _level="$1"
    local _prefix="$2"

    while read -r _line || [ -n "$_line" ]
    do
	bb_msg "$_level" "$_prefix$_line" || return "$?"
    done
}

# runs a command only if DRYRUN is undefined, emits it otherwise
function dw_run() {
    (( $# > 0 )) || {
	bb_msg err "common:dw_run called with no args"
	return
    }

    local _cmd="$1"
    shift

    if [ -z "${DRYRUN-}" ]
    then
	bb_msg debug "calling: $_cmd $*"
	"$_cmd" "$@" >/dev/null || return "$?"
    else
	bb_msg info "[dry run] $_cmd -n -v $*"

	[ -t 0 ] || {
	    local _input
	    if _input=$(cat)
	    then
		bb_msg_from_stdin info "<- " <<<"$_input"
	    fi
	}
	"$_cmd" -n -v "$@" 2>/dev/null | bb_msg_from_stdin info "-> " || \
	    return "$?"
    fi
}

# ldap_cmd read|write <ldaptools-executable> <arguments> [..]
# write commands are only done if no dry run
function ldap_cmd() {
    (( $# > 1 )) || \
	bb_fatal "common:ldap_cmd called with $# instead of 2 or more args"

    local _mode="$1"
    local _cmd="$2"
    shift 2
    local -a _args=("-Q" "-o" "ldif-wrap=no") # default ldap args

    if [[ -v LDAP_EXTRA_ARGS ]]
    then
	_args+=("${LDAP_EXTRA_ARGS[@]}")
    fi

    [ -z "${LDAP_URI-}" ] || _args+=("-H" "$LDAP_URI")
    [ -z "${LDAP_BINDDN-}" ] || _args+=("-D" "$LDAP_BINDDN")
    [ -z "${LDAP_LOCAL_ARGS[@]-}" ] || _args+=("${LDAP_LOCAL_ARGS[@]}")

    case "$_mode" in
	write) dw_run "$_cmd" "${_args[@]}" "$@";;
	read)
	    bb_msg debug "calling: $_cmd ${_args[*]} $*"
	    "$_cmd" "${_args[@]}" "$@" || return "$?"
	    ;;
	*) bb_fatal "common:ldap_cmd called with illegal mode: $_mode"
    esac
}

# config_init
# loads config file
function config_init() {
    (( $# == 0 )) || \
	bb_fatal "common:config_init called with $# instead of 0 args"

    [ -r "$CONFIGFILE" ] || \
	bb_fatal "${0##*/}: could not find config file $CONFIGFILE"

    bini_db_load config "$CONFIGFILE" || \
	bb_fatal "failed to load config: $CONFIGFILE"

    config_read_common "${SECTION-}"
}

# config_section_exists <section>
config_section_exists() {
    (( $# == 1 )) || \
	bb_fatal "common:config_section_exists called with $# instead of 1 arg"

    bini_section_exists config "$1"
}

# config_read <outvar> <key> [ <section> ]
# if section is specified but empty, it is ignored
function config_read() {
    (( $# == 2 || $# == 3 )) || \
	bb_fatal "common:config_read called with $# instad of 2 or 3 args"

    if [ -z "${3-}" ]
    then
	bini_get_value "$1" config "$2"
    else
	bini_get_value_with_default "$1" config "$2" "${3-}"
    fi
}

# config_read_optional <outvar> <key> [ <section> ]
function config_read_optional() {
    (( $# == 2 || $# == 3 )) || bb_fatal "common:config_read_optional" \
	"called with $# instad of 2 or 3 args"

    config_read "$1" "$2" "${3-}" || unset "$1"
}

# config_read_mandatory <outvar> <key> [ <section> ]
# returns 1 if config element is not found
function config_read_mandatory() {
    (( $# == 2 || $# == 3 )) || bb_fatal "common:config_read_mandatory" \
	"called with $# instad of 2 or 3 args"

    config_read "$1" "$2" "${3-}" || return 1
}

# config_read_mandatory <outvar> <key> <default> [ <section> ]
function config_read_with_default() {
    (( $# == 3 || $# == 4 )) || bb_fatal "common:config_read_with_default" \
	"called with $# instad of 3 or 4 args"

    config_read "$1" "$2" "${4-}" || eval "$1=\"\$3\""
}
# config_read_all [ <section> ]
# shellcheck disable=SC2034
function config_read_common() {
    (( $# < 2 )) || bb_fatal "common:config_read_common" \
	"called with $# instad of no or 1 arg"

    local SECTION="${1-}"

    # LDAP access
    config_read_optional LDAP_URI "ldap uri" "$SECTION"
    config_read_optional LDAP_BINDDN "ldap bind dn" "$SECTION"
    config_read_optional LDAP_EXTRA_ARGS "ldap extra args" "$SECTION"

    # LDAP tree config
    config_read_mandatory LDAP_SEARCH_BASE "ldap search base" "$SECTION"
    config_read_optional LDAP_USER_SUBHIER "ldap user subhier" "$SECTION"
    config_read_optional LDAP_GROUP_SUBHIER "ldap group subhier" "$SECTION"
    config_read_optional LDAP_AUTOFS_SUBHIER "ldap autofs subhier" "$SECTION"
    config_read_optional LDAP_AUTOFS_HOMES "ldap autofs homes" "$SECTION"

    # behaviour
    config_read WINDOWS_USERS "windows users" "$SECTION" || WINDOWS_USERS=false
    config_read AUTOFS_USERS "windows users" "$SECTION" || WINDOWS_USERS=false
    config_read_optional AUTOFS_HOMES_PREFIX "autofs homes prefix" "$SECTION"

    # homedirs
    config_read_optional HOMEDIR_HOST_EXTRA_PATHS "homedir host extra paths" \
	"$SECTION"
    config_read_optional HOMEDIR_BACKUP_PATH "homedir host backup path" \
	"$SECTION"
}

# launch cmd [ args [..] ]
#  launches other command of ldapusermgmt in same location
#  and honors common args
function lum_launch() {
    (( $# > 1 )) || \
	bb_fatal "common:lum_launch called with $# instad of 2 or more args"

    local _cmd="$1"
    shift

    bb_msg debug "launching $_cmd ${COMMON_LUM_ARGS[*]-} $*"
    if (( ${#COMMON_LUM_ARGS[@]} == 0 ))
    then
	"$_cmd" "$@"
    else
	"$_cmd" "${COMMON_LUM_ARGS[@]}" "$@"
    fi
}

# parse_common_args <main-callback> <custom-optstring> <op-callback>
#                   [ <args> [..]]
# parse common args and forward others to "custom-callback <char>"
# after parsing callback will be called with the remaining arguments
# note: if no custom-callback is needed call with two times ""
# shellcheck disable=SC2034
function parse_common_args() {
    (( $# > 2 )) || bb_fatal "common:parse_common_args" \
	"called with $# instead of 3 or more args"

    local _main_callback="$1"
    local _custom_opstring="${2-}"
    local _op_callback="${3-}"
    shift 3

    declare -g CONFIGFILE=/etc/ldapusermgmt/ldapusermgmt.conf
#    declare -g -a COMMON_LUM_ARGS=()
    local -a COMMON_LUM_ARGS=()

    # shellcheck disable=SC2119
    verbosity_init

    local _opchar
    while getopts ":c:Fhlns:qvV${_custom_opstring}" _opchar; do
	case "$_opchar" in
	    c) 
		declare -g CONFIGFILE="$OPTARG"; 
		COMMON_LUM_ARGS+=("-c" "$OPTARG")
		;;
	    F)
		declare -g FORCE=yes
		COMMON_LUM_ARGS+=("-F")
		;;
	    h) local_usage 0;;
	    s) 
		declare -g SECTION="$OPTARG";
		COMMON_LUM_ARGS+=("-S" "$OPTARG")
		;;
	    l)
		declare -g LDAP_ONLY=true
		COMMON_LUM_ARGS+=("-l")
		;;
	    n)
		# shellcheck disable=SC2034
		declare -g DRYRUN=true
		COMMON_LUM_ARGS+=("-n")
		;;
	    q)
		verbosity_set "$((VERBOSITY - 1))"
		COMMON_LUM_ARGS+=("-q")
		;;
	    v) 
		verbosity_set "$((VERBOSITY + 1))"
		COMMON_LUM_ARGS+=("-v")
		;;
	    V)
		echo "${0##*/}: version 0.2"
		bb_quit 0
		;;
	    *)

		[ "$_opchar" != '?' ] || _opchar="$OPTARG"

		if [ -z "$_op_callback" ]
		then
		    local_usage 1 "Unknown argument: $_opchar"
		else
		    "$_op_callback" "$_opchar" || \
			local_usage 1 "Unknown argument: $_opchar"
		fi
	esac
    done
    shift $((OPTIND - 1))

    "$_main_callback" "$@"
}

function load_config() {
    (( $# == 0 )) || \
	bb_fatal "common:load_config called with $# instad of no args"

    [ -f "$CONFIGFILE" ] || \
	bb_fatal "${0##*/}: could not find config file $CONFIGFILE"
    bini_db_load config "$CONFIGFILE" || \
	bb_fatal "failed to load config: $CONFIGFILE"

    config_read_common "${SECTION-}"
}

# files_helper <host> <operation> [ <var> .. ]
# remote host will be read from config (from $SECTION if not empty) unless
# the variable FILES_HELPER_HOST is defined
# returns 1 (unable to launch files_helper), 2 (files_helper failed)
function files_helper() {
    (( $# > 0 )) || \
	bb_fatal "common:files_helper called with $# instad of 1 or more args"

    local _host="$1"
    local _operation="$2"
    shift 2

    local _helper_pid
    if [ "$_host" == "localhost" ]
    then
	# no SSH
	_caller="/usr/sbin/lumfileshelper"
	[ -x "$_caller" ] || {
	    bb_msg err "unable to call files_helper on localhost," \
		"file not found: $_caller"
	    return 1
	}

	bb_msg debug "common:files_helper calling local lumfileshelper"
	if coproc helperproc {
		/usr/libexec/ldapusermgmt/lumfileshelper
	    }
	then
	    _helper_pid="$!"
	else
	    bb_msg err "common:files_helper failed to start local coproc"
	    return 1
	fi
    else
	# SSH
	local files_helper__user
	config_read_with_default files_helper__user "homedir host user" "root" \
	    "${SECTION-}"

	bb_msg debug "calling lumfileshelper via ssh" \
	    "$files_helper__user"@"$_host"
	if coproc helperproc {
		ssh -o BatchMode=yes \
		    -o ConnectTimeout=50 \
		    "$files_helper__user"@"$_host" \
		    /usr/libexec/ldapusermgmt/lumfileshelper
	    }
	then
	    _helper_pid="$!"
	else
	    bb_msg err "common:files_helper failed to start remote coproc"
	    return 1
	fi
    fi

    local _varname
    for _varname in VERBOSITY FORCE DRYRUN "$@"
    do
	[ -v "$_varname" ] || continue
	bb_msg debug "passing $(declare -p "$_varname")"
	declare -p "$_varname" >&"${helperproc[1]}"
    done

    if echo "start $_operation" >&"${helperproc[1]}"
    then
	# read helper stdout
	local _line
	while read -u "${helperproc[0]}" _line
	do
	    bb_msg warning "$_line"
	done
    else
	bb_msg err "failed to start files_helper process"
    fi

    wait "$_helper_pid" || {
	bb_msg err "files_helper failed with error code: $?"
	return 2
    }
}

# ldap_extract_attributes { <attribute-name> <retvn-array> } [..]
function ldap_extract_attributes() {
    (( $# % 2 == 0 )) || bb_fatal "ldap_extract_attributes called with $#" \
	"instead of an even number of args"

    local -A _retvns=() # one retvn-array for each attribute-name
    while (( $# > 1 ))
    do
	_retvns["$1"]="$2"
	shift 2
    done

    local _line _attr _value
    local -i _i=-1

    while read _line
    do
	if [[ "$_line" =~ ^[[:space:]]*$ ]]
	then
	    continue
	fi

	[[ "$_line" =~ : ]] || {
	    bb_msg err "ldap_extract_attributes: got illegal line: $_line"
	    continue
	}

	_attr="${_line%%:*}"
	_value="${_line#*: }"

	[[ "$_attr" != "dn" ]] || _i=$((_i + 1))

        [ -n "${_retvns[$_attr]-}" ] || continue

	eval "${_retvns[$_attr]}[$_i]=\"$_value\""
    done
}

# ldap_resolv_user_dns <retvn-array> <user-cn>
# store user-cn related dns in retvn-array
# returns 1 if none found
function ldap_resolv_user_dns() {
    (( $# == 2 )) || \
	bb_fatal "common:ldap_resolv_user_dns expects 2 but got $# args"

    local _ret_vn="$1"
    local _user_cn="$2"
    local -i _count=0

    bb_validate_vn ldap_resolv_user_dns "$_ret_vn"

    local _subhier _dn
    for _subhier in "${LDAP_USER_SUBHIER}" "${LDAP_GROUP_SUBHIER}" \
	"${LDAP_AUTOFS_HOMES},${LDAP_AUTOFS_SUBHIER}"
    do
	_dn=$(ldap_cmd read ldapsearch -LLL \
	    -b "$_subhier,${LDAP_SEARCH_BASE}" \
	    "cn=$_user_cn" "dn") || {
	    bb_msg err "failed to: ldapsearch -b " \
		"\"$_subhier,${LDAP_SEARCH_BASE}\" \"uid=$_user_cn\" \"dn\""
	    continue
	}

	[ -z "$_dn" ] || {
	    eval "$_ret_vn+=(\"${_dn#dn: }\")"
	    let _count++
	}
    done

    (( _count > 0 ))
}

# ldap_resolv_user_dns <retvn-array> <base>
function ldap_find_user_cns() {
    (( $# == 2 )) || \
	bb_fatal "ldap_find_user_cns called with $# instead of 2 args"

    local _ret_vn="$1"
    local _search_base="$2"

    bb_validate_vn ldap_find_user_cns "$_ret_vn"

    ldap_cmd read ldapsearch -LLL -b "$_search_base" cn
}

# find_user_homedir <ret-vn> <user-cn>
# returns users homedir or if autofs backed server:homedir
function ldap_resolv_user_homedir() {
    (( $# == 2 )) || \
	bb_fatal "common:ldap_resolv_user_homedir expects 2 but got $# args"

    local _ret_vn="$1"
    local _user_cn="$2"

    bb_validate_vn ldap_resolv_user_homedir "$_ret_vn"

    local _homedir
    local _search_base
    if [ -z "${LDAP_AUTOFS_SUBHIER-}" ]
    then
	printf -v _search_base "%s,%s" "${LDAP_AUTOFS_HOMES}" \
	    "${LDAP_SEARCH_BASE}"
    else
	printf -v _search_base "%s,%s,%s" "${LDAP_AUTOFS_HOMES}" \
	    "${LDAP_AUTOFS_SUBHIER}" "${LDAP_SEARCH_BASE}"
    fi
    _homedir=$(ldap_cmd read ldapsearch -LLL -b "$_search_base" \
	"cn=$_user_cn" "nisMapEntry" | \
	grep -E '^nisMapEntry: ') || {
	bb_msg debug "failed to ldap_cmd read ldapsearch -b \"$_search_base\"" \
	    "\"cn=$_user_cn\" \"nisMapEntry\""
	return 1
    }

    printf -v "$_ret_vn" "%s" "${_homedir#nisMapEntry: }"
}
